// ------------------------------------------------------------------------------
// http://www.arkitus.com
// Copyright (C) 2006  Ali Eslami
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// ------------------------------------------------------------------------------
// Game.java
// The game class. It will handle the user's inputs, run the AI, display graphics
// and basically do everything that isn't related to Processing's internal
// functions in AI.java
// ------------------------------------------------------------------------------

class Game
{
  Settings settings;              // Game settings loaded from the settings file
  GameWorld currentWorld;         // The actual game. Everything dynamic happens here
  Renderer renderer;              // Draws everything onto the screen
  boolean[] keys;                 // Keeps track of the keys that are pressed
  
  /**
   * Constructor
   */
  Game()
  {
    // Load settings
    settings = new Settings("settings.ais");
    
    // Load game world
    currentWorld = loadWorld("mainworld");
    
    // Initiate renderer
    renderer = new Renderer(settings);
    
    // Initiate keys
    keys = new boolean[5];
  }
  
  /**
   * Does everything the game should do
   * in one frame. Handle AI, display graphics
   * and so on.
   */
  void frame()
  {
    currentWorld.frame(keys);
    renderer.frame(currentWorld);
  }  
  
  /**
   * Handles mouse clicks
   */
  void mousePressed()
  {
  }
  
  /**
   * Handles mouse drags
   */
  void mouseDragged()
  {
  }
  
  /**
   * Handles key presses
   */
  void keyPressed()
  {
    if (keyCode == UP)
    {
      keys[0] = true;
    }
    if (keyCode == DOWN)
    {
      keys[1] = true;
    }
    if (keyCode == LEFT)
    {
      keys[2] = true;
    }
    if (keyCode == RIGHT)
    {
      keys[3] = true;
    }   
  }
  
  /**
   * Handles key releases
   */
  void keyReleased()
  {
    if (keyCode == UP)
    {
      keys[0] = false;
    }
    if (keyCode == DOWN)
    {
      keys[1] = false;
    }
    if (keyCode == LEFT)
    {
      keys[2] = false;
    }
    if (keyCode == RIGHT)
    {
      keys[3] = false;
    }  
  }
  
  /**
   * Loads a world and starts playing it.
   */
  GameWorld loadWorld(String worldName)
  {
    // The world that gets returned
    GameWorld R = new GameWorld();
    
    // An instance of the DataLoader class
    DataLoader dataLoader = new DataLoader();
    
    // Load gameworld data
    String[][] gameWorldData = dataLoader.loadData(settings.GAMEWORLD_PATH + worldName + ".aiw", "\\.");
    
    for (int i=0; i<gameWorldData.length; i++)
    {
      // gameWorldData[i][0]: Model name
      // gameWorldData[i][1]: Model class
      // gameWorldData[i][2]: Model polygon type
      // gameWorldData[i][3]: Model material
      // gameWorldData[i][4]: Model x position
      // gameWorldData[i][5]: Model y position
      // gameWorldData[i][6]: Model rotation
      // gameWorldData[i][7]: Model arguments
      
      // Load the model's polygon
      Polygon polygon = dataLoader.loadPolygon(gameWorldData[i][2]);
      
      // Load the model's material
      Material material = dataLoader.loadMaterial(gameWorldData[i][3]);
      
      // Create the resulting model and
      // add it to the gameworld.
      if (gameWorldData[i][1].equals("camera"))
      {
        R.addModel(new Camera(gameWorldData[i][0], polygon, material, new Point(Str2Int(gameWorldData[i][4]), Str2Int(gameWorldData[i][5])), radians(Str2Int(gameWorldData[i][6])), settings, gameWorldData[i][7]));
      }
      if (gameWorldData[i][1].equals("drag"))
      {
        R.addModel(new Drag(gameWorldData[i][0], polygon, material, new Point(Str2Int(gameWorldData[i][4]), Str2Int(gameWorldData[i][5])), radians(Str2Int(gameWorldData[i][6])), settings, gameWorldData[i][7]));
      }
      if (gameWorldData[i][1].equals("static"))
      {
        R.addModel(new Static(gameWorldData[i][0], polygon, material, new Point(Str2Int(gameWorldData[i][4]), Str2Int(gameWorldData[i][5])), radians(Str2Int(gameWorldData[i][6])), settings, gameWorldData[i][7]));
      }
      if (gameWorldData[i][1].equals("humanagent"))
      {
        R.addModel(new HumanAgent(gameWorldData[i][0], polygon, material, new Point(Str2Int(gameWorldData[i][4]), Str2Int(gameWorldData[i][5])), radians(Str2Int(gameWorldData[i][6])), settings, gameWorldData[i][7]));
      }
      if (gameWorldData[i][1].equals("sheepdispenser"))
      {
        R.addModel(new SheepDispenser(gameWorldData[i][0], polygon, material, new Point(Str2Int(gameWorldData[i][4]), Str2Int(gameWorldData[i][5])), radians(Str2Int(gameWorldData[i][6])), settings, gameWorldData[i][7]));
      }
      if (gameWorldData[i][1].equals("sheep"))
      {
        R.addModel(new Sheep(gameWorldData[i][0], polygon, material, new Point(Str2Int(gameWorldData[i][4]), Str2Int(gameWorldData[i][5])), radians(Str2Int(gameWorldData[i][6])), settings, gameWorldData[i][7]));
      }
    }
    
    return R;
  }
  
  /**
   * Convert a String to an Integer
   */
  int Str2Int(String S)
  {
    return Integer.parseInt(S);
  }
  
  /**
   * Convert a String to a Float
   */
  float Str2Flt(String S)
  {
    return (float) Integer.parseInt(S);
  }
}
