// ------------------------------------------------------------------------------
// http://www.arkitus.com
// Copyright (C) 2006  Ali Eslami
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// ------------------------------------------------------------------------------
// GameWorld.java
// The world. This will store all the models, update the models and so on.
// ------------------------------------------------------------------------------

class GameWorld
{
  ArrayList models;  
  
  /**
   * Constructor
   */
  GameWorld()
  {
    models = new ArrayList();
  }
  
  /**
   * Adds a model to the game world
   */
  void addModel(Model M)
  {
    models.add(M);
  }
  
  /**
   * Executes one frame of the game world
   */
  void frame(boolean[] keys)
  {
    for (int i=0; i<models.size(); i++)
    {
      // Update each model
      Model modelI = ((Model) models.get(i)).frame(models, keys);
      models.set(i, modelI);
      
      // Ask each model if it wants to
      // add anything to the world
      Model modelToBeAdded = modelI.modelToBeAdded();
      if (modelToBeAdded != null) addModel(modelToBeAdded);      
      
      if (modelI instanceof PhysicsModel)
      {
        for (int j=i-1; j>=0; j--)
        {
          Model modelJ = (Model) models.get(j);
          
          if (modelJ instanceof PhysicsModel)
          {
            collide((PhysicsModel) modelI, (PhysicsModel) modelJ);
          }
        }
      }
    }
  }
  
  /**
   * Detects and handles collisions
   * between two models
   */
  void collide(PhysicsModel A, PhysicsModel B)
  {
    float delta_x = A.center.x - B.center.x;
    float delta_y = A.center.y - B.center.y;
    float delta_d = sqrt(pow(delta_x,2)+pow(delta_y,2));
    float sum_r = A.poly.radius + B.poly.radius;
    
    if ((A.material.invDensity != 0 || B.material.invDensity != 0) && delta_d <= sum_r && (A.veloc.length() > 0 || B.veloc.length() > 0))
    {
      Vector MTD = calculateMTD(A, B);
      
      if (MTD.length() != 0)
      {
        Point[] contactPoints = findContactPoints(A, B, MTD);
        
        Vector vA = A.veloc;
        Vector vB = B.veloc;
        Vector vAB = vA.add(vB.multiply(-1));
        Vector rA = new Vector(contactPoints[0]);
        Vector rB = new Vector(contactPoints[1]);        
        
        float e = A.material.bounce*B.material.bounce;
        float invMassA = 1/A.mass;
        float invMassB = 1/B.mass;
        
        Vector vtA = A.veloc.add(rA.multiply(A.rotationSpeed));
        Vector vtB = B.veloc.add(rB.multiply(B.rotationSpeed));
        Vector vtAB = vA.add(vB.multiply(-1));
        
        float j = (-(1+e)*vAB.dot(MTD))/(MTD.dot(MTD)*(invMassA+invMassB)+pow(rA.normal().dot(MTD),2)*A.invMomentOfInertia+pow(rB.normal().dot(MTD),2)*B.invMomentOfInertia);

        if (invMassA != 0) A.impulse(MTD, 1 - A.mass/(A.mass + B.mass), +j, rA);
        if (invMassB != 0) B.impulse(MTD, B.mass/(A.mass + B.mass) - 1, -j, rB);
      }
    }
  }
  
  /**
   * Finds the points of contact
   * between the two objects
   */
  Point[] findContactPoints(PhysicsModel A, PhysicsModel B, Vector MTD)
  {
    Point[] R = new Point[2];
    R[0] = new Point();
    R[1] = new Point();
    
    Point[] supportA = findSupportPoints(A.poly, MTD);
    Point[] supportB = findSupportPoints(B.poly, MTD.multiply(-1));
    
    int sNumA = supportA.length;
    int sNumB = supportB.length;
    
    if (sNumA == 1 && sNumB == 1)
    {
      R[0] = supportA[0];
      R[1] = supportB[0];
    }
    else if (sNumA == 1 && sNumB == 2)
    {
      R[0] = supportA[0];
      Vector edge = new Vector(supportB[1].x-supportB[0].x, supportB[1].y-supportB[0].y);
      R[1] = new Point(supportA[0].x + A.center.x - B.center.x, supportA[0].y + A.center.y - B.center.y);
    }
    else if (sNumA == 2 && sNumB == 1)
    {
      R[1] = supportB[0];
      Vector edge = new Vector(supportA[1].x-supportA[0].x, supportA[1].y-supportA[0].y);
      R[0] = new Point(supportB[0].x + B.center.x - A.center.x, supportB[0].y + B.center.y - A.center.y);
    }
    
    return R;
  }
  
  /**
   * Finds the support points of the two objects
   * which is then used to calculate the contact
   * points
   */
  Point[] findSupportPoints(Polygon P, Vector V)
  {
    ArrayList R = new ArrayList(2);
    float minimumDot = new Vector(P.vertices[0]).dot(V);
    float[] dot = new float[P.vertices.length];
    dot[0] = minimumDot;
    
    for (int i=1; i<P.vertices.length; i++)
    {
      dot[i] = new Vector(P.vertices[i]).dot(V);
      if (dot[i] < minimumDot) minimumDot = dot[i];
    }

    final float threshold = 5E-10;

    for (int i=0; i<P.vertices.length; i++)
    {
      if (dot[i]-minimumDot < threshold)
      {
        R.add(P.vertices[i]);
        if (R.size() == 2) break;
      }
    }
    
    R.trimToSize();

    return (Point[]) R.toArray(new Point[R.size()]);
  }
  
  /**
   * Calculates the Mean 
   * Transitional Distance (?)
   * vector
   */
  Vector calculateMTD(PhysicsModel A, PhysicsModel B)
  {
    Vector newPushVector = new Vector();
    ArrayList pushVectors = new ArrayList();

    for (int j=0; j<A.poly.numEdges(); j++)
    {
      float intervalIntersection = calculateIntervalIntersection(A.poly.getEdge(j).normal(), A, B);
      if (intervalIntersection == 0) return new Vector();
      newPushVector = new Vector();
      newPushVector.set(intervalIntersection, A.poly.getEdge(j).normal().dir());
      pushVectors.add(newPushVector);
    }
    for (int j=0; j<B.poly.numEdges(); j++)
    {       
      float intervalIntersection = calculateIntervalIntersection(B.poly.getEdge(j).normal(), A, B);
      if (intervalIntersection == 0) return new Vector();
      newPushVector = new Vector();
      newPushVector.set(intervalIntersection, B.poly.getEdge(j).normal().dir());
      pushVectors.add(newPushVector);
    }

    Vector currentBest = new Vector();

    for (int i=0; i<pushVectors.size(); i++)
    {
      Vector pushVector = (Vector) pushVectors.get(i);
            
      if ((pushVector.length() < currentBest.length()) || currentBest.length() == 0)
      {
        currentBest = pushVector;
      }
    }

    Vector D = new Vector();
    D.a = A.center.x - B.center.x;
    D.b = A.center.y - B.center.y;
    
    if (D.dot(currentBest) < 0) currentBest = currentBest.multiply(-1);

    return currentBest; 
  }
  
  /**
   * Calculates the intersecting
   * interval of two objects in
   * a certain direction
   */
  float calculateIntervalIntersection(Vector N, PhysicsModel A, PhysicsModel B)
  {
    Interval intervalA = calculateInterval(N,A);
    Interval intervalB = calculateInterval(N,B);
    Interval intersection = intervalA.intersect(intervalB); 
    return intersection.length();
  }
  
  /**
   * Projects the object onto an axis
   */
  Interval calculateInterval(Vector axis, PhysicsModel O)
  { 
    Interval R = new Interval(0,0);
    for (int i=0; i<O.poly.numEdges(); i++)
    {
      R.insert((O.poly.getEdge(i).dot(axis) + new Vector(O.poly.vertices[i]).dot(axis))/axis.length());
    }
    return R.offset((new Vector(O.center).dot(axis))/axis.length());
  } 
}
