// ------------------------------------------------------------------------------
// http://www.arkitus.com
// Copyright (C) 2006  Ali Eslami
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// ------------------------------------------------------------------------------
// Sheep.java
// The sheep that the human plays with.
// ------------------------------------------------------------------------------

class Sheep extends PhysicsModel
{
  float initVelocRange;
  float neighbourRadius;
  float separateK;
  float stickK;
  float neighbourSeparateRadius;
  
  Vector steer;
  
  /**
   * Constructor
   */
  Sheep(String N, Polygon P, Material M, Point C, float R, Settings S, String I)
  {
    super(N, "Sheep", P, M, C, R, S, I);
    
    veloc = new Vector(random(-initVelocRange/2,initVelocRange/2), random(-initVelocRange/2,initVelocRange/2));
    steer = new Vector();
  }
  
  /**
   * Executes one frame
   */
  Model frame(ArrayList models, boolean[] keys)
  {
    reset();
        
    ArrayList neighbours = findNeighbours(models);
    
    if (neighbours.size() > 0) {
      // align(neighbours);
      stick(neighbours);
      separate(neighbours);
      escape(neighbours);
      wander();
    }
    else {
      follow(models);
    }
    
    steer();
    move();
    damp(0.98, 0.90);
    return this;
  }
  
  /**
   * Updates the sheep's velocity using
   * the steer vector
   */
  void steer()
  {
    Point head = new Point(0,10).rotate(poly.rotation);
    Force steerForce = new Force(head, steer);
    forces.add(steerForce);
  }
  
  /**
   * Finds its neighbours in the list of models
   * in the world. Ideally has to be very fast.
   */
  ArrayList findNeighbours(ArrayList models)
  {
    ArrayList R = new ArrayList();
    
    for (int i=0; i<models.size(); i++)
    {
      Model model = (Model) models.get(i);
      float d2 = pow(center.x-model.center.x,2)+pow(center.y-model.center.y,2);
      
      if ((model.name.equals("sheep") || model.name.equals("humanagent")) && d2 != 0 && d2 < pow(neighbourRadius,2))
      {
        R.add(model);
      }
    }
    
    return R;
  }
  
  /**
   * Tries to avoid colliding with any of its neighbours
   */
  void separate(ArrayList models)
  {
    if (models.size() != 0)
    {
      int count = 0;
      
      // Find the mean center of all the
      // points in its neighbourhood
      Point meanCenter = new Point();
      for (int i=0; i<models.size(); i++)
      {  
        Model model = (Model) models.get(i);
        float d2 = pow(center.x-model.center.x,2)+pow(center.y-model.center.y,2);
        
        if (d2 < pow(neighbourSeparateRadius,2)) 
        {
          meanCenter = new Point(meanCenter.x + model.center.x, meanCenter.y + model.center.y);
          count++;
        }
      }
      
      if (count != 0)
      {
        meanCenter = meanCenter.multiply(1/((float) count));
      
        // RUN AWAY!
        Vector d = new Vector(center.x-meanCenter.x, center.y-meanCenter.y);
        steer = steer.add(d.multiply(separateK/100));
      }      
    }
  }
  
  /**
   * Run from the human agent
   */
  void escape(ArrayList models)
  {
    if (models.size() != 0)
    {
      for (int i=0; i<models.size(); i++)
      {  
        Model model = (Model) models.get(i);
        
        if (model.name.equals("humanagent"))
        {
          // RUN AWAY!
          Vector d = new Vector(center.x-model.center.x, center.y-model.center.y);
          steer = steer.add(d.multiply(0.1));   
        }
      }
    }
  }
  
  /**
   * Follow the human agent
   */
  void follow(ArrayList models)
  {
    if (models.size() != 0)
    {
      for (int i=0; i<models.size(); i++)
      {  
        Model model = (Model) models.get(i);
        
        if (model.name.equals("humanagent"))
        {
          // Follow the agent
          Vector d = new Vector(center.x-model.center.x, center.y-model.center.y);
          steer = steer.add(d.multiply(-0.02));   
        }
      }
    }
  }
  
  /**
   * Makes the sheep stick to their neighbours
   */
  void stick(ArrayList models)
  {
    if (models.size() != 0)
    {
      // Find the mean center of all the
      // points in its neighbourhood
      Point meanCenter = new Point();
      for (int i=0; i<models.size(); i++)
      {  
        Model model = (Model) models.get(i);
        meanCenter = new Point(meanCenter.x + model.center.x, meanCenter.y + model.center.y);
      }
      meanCenter = meanCenter.multiply(1/((float) models.size()));
      
      // STICK!
      Vector d = new Vector(meanCenter.x-center.x, meanCenter.y-center.y);
      steer = steer.add(d.multiply(stickK/100));
    }
  }
  
  /**
   * Just wander around the map
   */
  void wander()
  {
    Vector deviation = veloc.normal().multiply(random(-0.1,0.1));
    steer = steer.add(veloc.add(deviation).multiply(0.035));
  }
   
  /**
   * Parse the input string
   */
  void parseInput(String input)
  {
    String[] args = inputToArray(input);
     
    initVelocRange = Str2Int(args[0]);
    neighbourRadius = Str2Flt(args[1]);
    neighbourSeparateRadius = Str2Flt(args[2]);
    separateK = Str2Flt(args[3]);
    stickK = Str2Flt(args[4]);
  }
  
  /**
   * Resets values to get ready for the 
   * next frame
   */
  void reset()
  {
    steer = new Vector();
    forces = new ArrayList();
    accel = new Vector();
    torque = 0; 
  }
}
  
